/*
 *	Hacker.c
 *
 *	Written: February 10, 1982 by Stoney Ballard
 *
 *	Edit History:
 *
 *	This program reads in the VAX/Smalltalk-80 .VMI file and then
 *	calls a user-written routine to operate on it.
 *
 *
 *
 *
 *			DEC Virtual Memory Image Layout
 *		 
 *			   Format for a .VMI file
 *		
 *		
 *		block				    byte offset
 *			+-----------------------+
 *		 1	|  Size of OM in bytes	| :0
 *			+-----------------------+
 *			|  Size of OT in bytes  | :4
 *			+-----------------------+
 *			|  Static byte length	| :8
 *			+-----------------------+
 *			|  Static object count	| :12
 *			+-----------------------+
 *			|     Zero filled	| :16-511
 *			+-----------------------+
 *		 2	| 			|
 *			|     Static Region	|
 *			|			|
 *			+-----------------------+
 *			|			|
 *			|    Dynamic Region	|
 *			|			|
 *			+-----------------------+
 *			|			|
 *			|    Object Table	|
 *			|			|
 *			+-----------------------+

 *			  Object Table Entry
 *			  ------------------
 *
 *       15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *      |                       Location (low)                  |   | F |
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *
 *       31  30  29  28  27  26  25  24  23  22  21  20  19  81  17  16
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *      | S |NE | W | C | M |S2 |S1 | P |         Location (high)       |
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *
 *  F:  Free entry
 *  P:  Pointers in this object
 *  S1: Swept-1
 *  S2: Swept-2
 *  M:  Mark bit
 *  C:  Cached
 *  W:  Object has words
 *  NE: Object needs entry
 *  S:  Object is static
 *  Location:  24 bit pointer to the physical address of the object (always
 *	       longword aligned); points to the size field of the object

 *  Object Format
 *  -------------
 *  Note:  All fields are longwords.
 *
 *               +----------------+
 *  Location --> |     size       |
 *               +----------------+
 *               |     class      |
 *               +----------------+
 *		 |    field 1     |
 *               +----------------+
 *               |    field 2     |
 *               +----------------+
 *               |                |
 *                       .
 *                       .
 *                       .
 *               |                |
 *               +----------------+
 *               |    field n     |
 *               +----------------+
 *
 *  size:   size in bytes of whole object (including size & class field);
 *	    allocated in longword increments
 *  class:  oop pointer to class of object
 *  field:  slot for acquaintence
 *
 */

#include	<stdio.h>
#include	<ctype.h>

/* booleans */
#define TRUE 1
#define FALSE 0

/* maximum region sizes */
#define omMax		(1024*256*4)	/* max byte size of object space */
#define	otMax		(128*1024)	/* number of nodes in OT */

/* useful object names */
#define oopNil		002
#define oopFalse	004
#define oopTrue		006
#define oopProcessor	010
#define oopUstable	012
#define oopSmallInteger	014
#define oopString	016
#define oopArray	020
#define oopSmalltalk	022
#define	oopFloat	024
#define oopMethodCtx	026
#define oopBlockCtx	030
#define oopPoint	032
#define oopLargeInt	034
#define oopDispBitmap	036
#define oopMessage	040
#define	oopCompiled	042
#define oopSemaphore	046
#define oopCharacter	050
#define oopDoesNotUnd	052
#define oopCantReturn	054
#define oopMonitor	056
#define oopSpecMess	060
#define oopCharTable	062
#define oopMustBeBool	064

#define oopBitmap	010074
#define oopStrikeFont	0x0D0C
#define oopClass	0x46
#define oopSymbol	0x38
#define oopMetaClass	0x3C

/* bits in ote */
#define	oteAddress	0x00FFFFFE	/* mask for address */
#define oteFree		0x00000001
#define otePointers	0x01000000
#define oteSwept1	0x02000000
#define oteSwept2	0x04000000
#define oteMark		0x08000000
#define oteCached	0x10000000
#define oteWords	0x20000000
#define oteNeedsEntry	0x40000000
#define oteStatic	0x80000000


/* object fields */
#define behavior_sclass	0

#define class_name	6

#define metaClass_class	6

#define font_xTable	0
#define font_glyphs	1
#define	font_name	2
#define font_stopCond	3
#define font_type	4
#define font_minAscii	5
#define font_maxAscii	6
#define font_maxWidth	7
#define font_strikeLen	8
#define font_ascent	9
#define font_descent	10
#define font_xOffset	11
#define font_raster	12
#define font_subscript	13
#define font_superscript 14
#define font_emphasis	15

#define form_bits	0
#define form_width	1
#define form_height	2
#define form_offset	3

/* file stuff */
FILE	*vmi;			/* iop for .VMI file */
FILE	*fontfile;		/* iop for font file */

/* Smalltalk image buffers */
int	header[128];		/* block buffer for head of .VMI file */
int	ot[otMax];		/* buffer for object table */
char	om[omMax];		/* buffer for object space */
int	somSize;		/* size in lwords of static object space */
int	somCount;		/* number of static objects */
int	domSize;		/* size in lwords of dynamic object space */
int	otSize;			/* size in lwords of object table */
int	maxoop;			/* max oop (/2) value */
char stringbuf[256];

short	bittable[16] = { 0x0001,	/* bit 0 */
			 0x0002,	/* bit 1 */
			 0x0004,	/* bit 2 */
			 0x0008,	/* bit 3 */
			 0x0010,	/* bit 4 */
			 0x0020,	/* bit 5 */
			 0x0040,	/* bit 6 */
			 0x0080,	/* bit 7 */
			 0x0100,	/* bit 8 */
			 0x0200,	/* bit 9 */
			 0x0400,	/* bit 10 */
			 0x0800,	/* bit 11 */
			 0x1000,	/* bit 12 */
			 0x2000,	/* bit 13 */
			 0x4000,	/* bit 14 */
			 0x8000	};	/* bit 15 */

main(argc, argv)

int	argc;
char	*argv[];

{
	char	*vmiFile;
	int	loc;
	int	i;

	vmiFile = argv[1];
	if (argc <= 1) error("?No file name given?\n");
	vmi = fopen(vmiFile, "r");
	if (vmi == NULL) error("Can't open input file\n");
	loadImage();
        printf("Load done\n");
	operate();
	printf("Completed hacking\n");
}

error(string)

char	*string;

{
	printf(string);
	exit();
}

loadImage()
{
  int	i, dbase, sbase, dyncount, statcount;
  char	*somptr, *domptr;

  printf("Loading image\n");
  if (fread(&header, 512, 1, vmi) == 0)
    error("Error reading file (header)\n");
  domSize = header[0];	/* dynamic object area byte length */
  otSize = header[1];	/* object table byte length */
  somSize = header[2];	/* static object area byte length */
  somCount = header[3];	/* static object count */
  somptr = &om;
  printf("Loading %d bytes of Statics\n", somSize);
  if (fread(somptr, 512, (somSize + 511) >> 9, vmi) == 0)
    error("Error reading file (som)\n");
  domptr = somptr + somSize;
  printf("Loading %d bytes of Dynamics\n", domSize);
  if (fread(domptr, 512, (domSize + 511) >> 9, vmi) == 0)
    error("Error reading file (dom)\n");
  printf("Loading %d bytes of Object Table\n", otSize);
  if (fread(&ot, 512, (otSize + 511) >> 9, vmi) == 0)
    error("Error reading file (ot)\n");
  printf("Load done\n");
  /* now, fixup the ot */
  maxoop = (otSize >> 2) - 1;
  printf("There are %d oops in the image\n", maxoop);
  dbase = domptr - &om;
  sbase = somptr - &om;
  dyncount = statcount = 0;
  for (i = 1; i <= maxoop; ++i) {
    if ((ot[i] & oteFree) == 0) {
      if ((ot[i] & oteStatic) == 0) {
        ot[i] = ot[i] + dbase;
	++dyncount;
      }
      else {
        ot[i] = ot[i] + sbase;
	++statcount;
      }
    }
  }
  maxoop = maxoop << 1;
  printf("There are %d static objects and %d dynamics\n", statcount,
    dyncount);
}

/* useful utilities */

/* returns 1 if oop is free */
int isfree(oop)
int oop;
{
  return(ot[oop >> 1] & oteFree);
}

/* returns the base address of an object */
int addrof(oop)
int oop;
{
  return(ot[oop >> 1] & oteAddress);
}

/* returns the object name from a field given byte offset */
int oopat(base, fieldnum)
int base, fieldnum;
{
  int *iptr;

  iptr = om + base + (fieldnum << 2);
  return(*iptr);
}

/* returns the 16 bit word from a word field given byte base offset */
int wordat(base, fieldnum)
int base, fieldnum;
{
  unsigned short *iptr;

  iptr = om + base + (fieldnum << 1);
  return(*iptr);
}


/* returns the nth field from an object  - skips header */
int fieldat(oop, field)
int oop, field;
{
  return(oopat(addrof(oop), field + 2));
}

/* returns an integer from a field */
intfieldat(oop, field)
int oop, field;
{
  return(fieldat(oop, field) >> 1);
}

/* returns the class of an object */
classof(oop)
int oop;
{
  if (oop & 1)
    return(oopSmallInteger);
  else
    return(oopat(addrof(oop), 1));
}

/* returns the superclass of an object */
superclassof(oop)
int oop;
{
  fieldat(oop, behavior_sclass);
}

/* returns the byte length of an object */
int lenof(oop)
int oop;
{
  return((oopat(addrof(oop), 0) - 8) & 0xFFFFFF);
}

/* checks whether an object is or inherits from the given class */
iskindof(oop, class)
int oop, class;
{
  int cl;

  cl = classof(oop);
  if (cl == class)
    return(TRUE);
  while (cl != oopNil) {
    cl = superclassof(cl);
    if (cl == class)
      return(TRUE);
  }
  return(FALSE);
}

/* tests whether an object is directly of the given class */
isa(oop, class)
int oop, class;
{
  return(class == classof(oop));
}

/* packs a smalltalk string into a c string buffer */
stringof(oop, sptr)
int oop;
char *sptr;
{
  int len, i, base, sclass;

  sclass = classof(oop);
  if (!iskindof(oop, oopString))
    strcpy(sptr, "?Not a String?");
  else {
    len = lenof(oop);
    base = addrof(oop) + 8;
    for (i = 0; i < len; ++i)
      *sptr++ = om[base++];
    *sptr = '\0';
  }
}

/* create the name of an object */
nameof(oop, sptr)
int oop;
char *sptr;
{
  if (iskindof(oop, oopMetaClass)) {
    strcpy(sptr, "Meta ");
    sptr += 5;
    stringof(fieldat(fieldat(oop, metaClass_class), class_name), sptr);
  }
  else if (iskindof(oop, oopClass)) {
    strcpy(sptr, "Class ");
    sptr += 6;
    stringof(fieldat(oop, class_name), sptr);
  }
  else {
    stringof(fieldat(classof(oop), class_name), sptr);
  }
}

/* operate on the image */
operate()
{
  int oop;

  printf("Operating\n");
  for (oop = 2; oop <= maxoop; oop += 2) {
    if (!isfree(oop)) {
      if (isa(oop, oopStrikeFont))
        writefont(oop);
    }
  }
}

/* writes out a font */
writefont(oop)
int oop;
{
  char fontfname[64];
  int rmsfontfile, glyphbitmap, xtable, height, raster, i, base;
  int minascii, maxascii, lc, glyphs;

  stringof(fieldat(oop, font_name), stringbuf);
  printf("For StrikeFont %s\n", stringbuf);
  printf("Enter file name: ");
  scanf("%s", fontfname);
  printf("File name is '%s'\n", fontfname);
  if (-1 == (rmsfontfile = creat(fontfname, 0,
       "dna = font.fnt", "rfm = var", "rat = cr"))) {
     printf("Can't open file '%s'\n", fontfname);
     return;
  }
  fontfile = fdopen(rmsfontfile, "w");
  printf("Writing parameters\n");
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_type));
  minascii = intfieldat(oop, font_minAscii);
  fprintf(fontfile, "%04x\n", minascii);
  maxascii = intfieldat(oop, font_maxAscii);
  fprintf(fontfile, "%04x\n", maxascii);
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_maxWidth));
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_strikeLen));
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_ascent));
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_descent));
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_xOffset));
  fprintf(fontfile, "%04x\n", intfieldat(oop, font_raster));
  /* now, write out the glyphs */
  /* format is 8 16bit words/line */
  printf("Writing glyphs\n");
  raster = intfieldat(oop, font_raster);
  height = intfieldat(oop, font_ascent) + intfieldat(oop, font_descent);
  glyphs = fieldat(oop, font_glyphs);
  glyphbitmap = fieldat(glyphs, form_bits);
  base = addrof(glyphbitmap) + 8;
  lc = 0;
  for (i = 0; i < (raster * height); ++i) {
    if (lc > 0)
      fprintf(fontfile, " ");
    fprintf(fontfile, "%04x", wordat(base, i));
    if (++lc >= 8) {
      fprintf(fontfile, "\n");
      lc = 0;
    }
  }
  if (lc > 0)
    fprintf(fontfile, "\n");
  /* now, write out the xTable */
  fprintf(fontfile, "*\n");
  printf("Writing xTable\n");
  xtable = fieldat(oop, font_xTable);
  lc = 0;
  for (i = minascii; i <= (maxascii + 2); ++i) {
    if (lc > 0)
      fprintf(fontfile, " ");
    fprintf(fontfile, "%04x", intfieldat(xtable, i));
    if (++lc >= 8) {
      fprintf(fontfile, "\n");
      lc = 0;
    }
  }
  fprintf(fontfile, "\n");
  /* done */
  fclose(fontfile);
}


