/*
 *
 * Module:	MAKER.C
 *
 * Created:	9-Mar-1981		Steven H. Gutfreund
 * Updated:	2-Sep-1981		Stephen F. Shirron
 * Updated:	9-Sep-1981		Stephen F. Shirron
 *
 * System:	Smalltalk-80
 *
 * Functional Description:
 *
 * This module converts a XEROX PARC format virtual memory image file into
 * a DEC .VMI file for loading a 11/23 Smalltalk machine. This program takes
 * one argument on the command line:  the name of the file to convert.  This
 * file should be a undefined RMS sequential file.  The program creates a .VMI
 * file of the same name.
 *
 *
 *			DEC Virtual Memory Image Layout
 *		 
 *			   Format for a .VMI file
 *		
 *		
 *		block				    size
 *			+-----------------------+
 *		 1	|  Size of OT in bytes	| longword (4 bytes)
 *			+-----------------------+
 *			|  Size of OB in bytes  | longword (4 bytes)
 *			+-----------------------+
 *			|     Zero filled	| 504 bytes
 *			+-----------------------+
 *		 2	| 			| 64K bytes (16K entries)
 *			|     Object Table	|
 *			|			|
 *			+-----------------------+
 *	       130	|			| ??
 *			|	Objects		|
 *			|			|
 *			+-----------------------+

 *			  Object Table Entry
 *			  ------------------
 *
 *       15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *      |                       Location (low)                  |   | F |
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *
 *       31  30  29  28  27  26  25  24  23  22  21  20  19  81  17  16
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *      |   |   | W | L | D | S | M | P |         Location (high)       |
 *      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
 *
 *  F:  Free entry
 *  P:  Pointers in this object
 *  M:  Mark bit
 *  S:  Swept bit
 *  D:  Dirty bit
 *  L:  Locked bit
 *  W:  Words in this object
 *  Location:  24 bit pointer to the physical address of the object (always
 *	       longword aligned); points to the size field of the object

 *  Object Format
 *  -------------
 *  Note:  All fields are longwords.
 *
 *               +----------------+
 *  Location --> |     size       |
 *               +----------------+
 *               |     class      |
 *               +----------------+
 *		 |    field 1     |
 *               +----------------+
 *               |    field 2     |
 *               +----------------+
 *               |                |
 *                       .
 *                       .
 *                       .
 *               |                |
 *               +----------------+
 *               |    field n     |
 *               +----------------+
 *
 *  size:   size in bytes of whole object (including size & class field);
 *	    allocated in longword increments
 *  class:  oop pointer to class of object
 *  field:  slot for acquaintence
 *
 */

#include	<stdio.h>
#include	<ctype.h>

#define	alloc		malloc

#define obMax		(1024*1024)	/* maximum size of object space */
#define	otMax		(1024*32*4)	/* number of nodes in OT
					   times 4 bytes/node */
#define	oopCompiled	042		/* oop for Class Compiled Method */
#define	oopFloat	024		/* oop for Class Float */
#define oopBitmap	010074		/* oop for Class Bitmap */
#define oopMethodCtx	026		/* oop for Class MethodCtx */
#define oopBlockCtx	030		/* oop for Class BlockCtx */
#define	oldLocation	0x000fffff	/* old location bits for object */
#define	oldFree		(1<<21)		/* OT entry is free */
#define	oldHasPointers	(1<<22)		/* object is made up of pointers */
#define	oldOdd		(1<<23)		/* object has odd length */
#define	newLocation	0x00fffffc	/* new location bits for object */
#define newFree		(1<<0)		/* OT entry is free */
#define newHasPointers	(1<<24)		/* object is made up of pointers */
#define newHasWords	(1<<29)		/* object has word fields */
#define	Nlits		0x00007e00	/* swabbed Nlits of compiled method */
#define	litsOffset	9		/* swabbed offset for Nlits */

FILE	*dec;			/* iop for .VMI file */
FILE	*parc;			/* iop for Xerox Parc file */
FILE	*changes;		/* iop for changes file */
int	*header;		/* block buffer for head of .VMI file */
int	*ot;			/* buffer holding ot */
short	*ob;			/* buffer holding objects */
int	newOT[otMax/4];		/* buffer for new object table */
char	newOB[obMax];		/* new object space */
char	decFile[80];		/* name of DEC .VMI file to create */
int	newOBpointer;		/* pointer into the object space */
int	obSize;			/* size in bytes of object space */
int	otSize;			/* size in bytes of object table */
short	bittable[16] = { 0x0001,	/* bit 0 */
			 0x0002,	/* bit 1 */
			 0x0004,	/* bit 2 */
			 0x0008,	/* bit 3 */
			 0x0010,	/* bit 4 */
			 0x0020,	/* bit 5 */
			 0x0040,	/* bit 6 */
			 0x0080,	/* bit 7 */
			 0x0100,	/* bit 8 */
			 0x0200,	/* bit 9 */
			 0x0400,	/* bit 10 */
			 0x0800,	/* bit 11 */
			 0x1000,	/* bit 12 */
			 0x2000,	/* bit 13 */
			 0x4000,	/* bit 14 */
			 0x8000	};	/* bit 15 */

main(argc, argv)

int	argc;
char	*argv[];

{
	char	*parcFile;
	char	*changesFile;
	int	loc;
	int	i;

	parcFile = argv[1];
	if (argc <= 1) error("?How can I make nothing?\n");
	loc = strlen(parcFile) - 1;
	for (; parcFile[loc] != '.'; loc--) ;
	strcpy(decFile,parcFile);
	strcpy(&decFile[loc], ".vmi");
	parc = fopen(parcFile, "r");
	if (parc == NULL) error("Can't open input file\n");
	dec = fopen(decFile, "w");
	if (dec == NULL) error("Can't make new file\n");
	changesFile = argv[2];
	if (argc <= 2) {
		printf("No changes being processed\n");
		changes = NULL; }
	else {
		changes = fopen(changesFile, "r");
		if (changes == NULL) error("Can't open changes file\n"); }
	setup();
	for (i = 1; i < otSize/4; i++) swabber(i);
	finish();
	exit();
}

error(string)

char	*string;

{
	printf(string);
	exit();
}

setup()
{
	int	i;

	header = alloc(512);
	fread(header, 512, 1, parc);
	swab(header, 8);
	wswab(header, 4);
	obSize = header[0] * 2;
	otSize = header[1] * 2;
	ob = alloc((obSize + 511) & ~511);
	fread(ob, 512, (obSize + 511) >> 9, parc);
	ot = alloc((otSize + 511) & ~511);
	fread(ot, 512, (otSize + 511) >> 9, parc);
	swab(ot, otSize);
	wswab(ot, otSize/2);
	for (i = 0; i < otMax/4; i++) newOT[i] = newFree;
	newOBpointer = 0;
}

swab(what, bytes)

char	*what;
int	bytes;

{
	int	i;
	char	temp;

	for (i = 0; i < bytes; i += 2) {
		temp = what[i+1];
		what[i+1] = what[i];
		what[i] = temp; }
}

wswab(what, words)

short	*what;
int	words;

{
	int	i;
	short	temp;

	for (i = 0; i < words; i += 2) {
		temp = what[i+1];
		what[i+1] = what[i];
		what[i] = temp; }
}

swabber(oop)

int	oop;

{
	int	i;
	int	j;
	short	temp;
	int	rwm;
	int	count;
	int	class;
	int	lits;
	int	memPointer;
	union {	int	longw[1];
		short	wordw[2];
		char	bytew[4]; } transfer;

	if (ot[oop] & oldFree) return;
	rwm = ot[oop] & oldLocation;
	swab(&ob[rwm], 4);		/* swap length and class fields */
	count = (ob[rwm++]*2)-4;	/* get object size in bytes */
	class = ob[rwm++];		/* and also get class oop */
	memPointer = newOBpointer;	/* save pointer for later */
	newOT[oop] = memPointer;	/* set up OTE */
	memPointer += 4;		/* save space for length */
	transfer.wordw[0] = class;	/* get class oop */
	transfer.wordw[1] = 0;		/* zero extend class oop */
	for (i = 0; i < 4; i++) newOB[memPointer++] = transfer.bytew[i];	

/*
 *	There are three cases to consider:
 *
 *	1)  "P" bit set -- all fields are words, so swap them
 *	2)  oopClass == CompiledMethod -- swap length, class, header,
 *	    and all literals
 *	3)  neither of the above -- swap only length and class fields
 */

	if (ot[oop] & oldHasPointers) {
		newOT[oop] |= newHasPointers;
		swab(&ob[rwm], count);
		for (i = 0; i < count; i += 2) {
			if (ob[rwm] & 1)
				transfer.longw[0] = ob[rwm++];
			else {
				transfer.wordw[0] = ob[rwm++];
				transfer.wordw[1] = 0; }
			for (j = 0; j < 4; j++)
				newOB[memPointer++] = transfer.bytew[j]; }}
	else if (class == oopCompiled) {
		lits = (ob[rwm] & Nlits) >> litsOffset;
		lits *= 2;		/* convert to a byte count */
		lits += 2;		/* don't forget the method header */
		swab(&ob[rwm], lits);
		for (i = 0; i < lits; i += 2) {
			if (ob[rwm] & 1)
				transfer.longw[0] = ob[rwm++];
			else {
				transfer.wordw[0] = ob[rwm++];
				transfer.wordw[1] = 0; }
			for (j = 0; j < 4; j++)
				newOB[memPointer++] = transfer.bytew[j]; }
		count -= lits;
		for (i = 0; i < count; i += 2) {
			transfer.wordw[0] = ob[rwm++];
			newOB[memPointer++] = transfer.bytew[0];
			newOB[memPointer++] = transfer.bytew[1]; }}
	else if ((class == oopFloat) || (class == oopBitmap)) {
		newOT[oop] |= newHasWords;
		for (i = 0; i < count; i += 2) {
			transfer.wordw[0] = ob[rwm++];
			if (class == oopBitmap) {
				temp = 0;
				for (j = 0; j < 16; j++) {
					if (transfer.wordw[0] & bittable[j])
						temp |= bittable[15-j]; }
				transfer.wordw[0] = temp; }
			newOB[memPointer++] = transfer.bytew[1];
			newOB[memPointer++] = transfer.bytew[0]; }
		if (class == oopFloat) toFloat(&newOB[memPointer-4]); }
	else for (i = 0; i < count; i += 2) {
			transfer.wordw[0] = ob[rwm++];
			newOB[memPointer++] = transfer.bytew[0];
			newOB[memPointer++] = transfer.bytew[1]; }
	if (ot[oop] & oldOdd) memPointer--;
	transfer.longw[0] = memPointer - newOBpointer;
	for (i = 0; i < 4; i++) newOB[newOBpointer++] = transfer.bytew[i];
	newOBpointer = (memPointer + 3) & newLocation;
}

toFloat(number)

float	*number;

{
	number[0] *= 4;
}

finish()

{
	int	i;
	int	j;
	int	k;
	int	mem;
	int	class;
	int	method;
	int	methptr;
	int	nlits;
	int	ip;
	int	home;
	int	homeptr;
	int	oop;
	int	field;
	int	new;
	char	temp;

	header[0] = newOBpointer;
	header[1] = otMax;
	printf("\nOTE's = %d\n",otSize/4);
	printf("Old Object Space = %d bytes\n",obSize);
	printf("New Object Space = %d bytes\n",newOBpointer);

	printf("\nScanning OTE for bad entries...\n");
	for (i = 1; i < otSize/4; i++) {
		mem = newOT[i] & newLocation;
		mem += 4;
		class = (newOB[mem] & 0x00ff) + ((newOB[mem+1]<<8) & 0xff00);
		if (class == oopMethodCtx) {
			method = (newOB[mem+16] & 0x00ff) +
				 ((newOB[mem+17]<<8) & 0xff00);
			methptr = newOT[method/2] & newLocation;
			nlits = newOB[methptr+8] & 0x007e;
			ip = (newOB[mem+8] & 0x00ff) +
			     ((newOB[mem+9]<<8) & 0xff00);
			ip += (nlits+2)<<1;
			newOB[mem+8] = ip & 0xff;
			newOB[mem+9] = (ip>>8) & 0xff; }
		else if (class == oopBlockCtx) {
			home = (newOB[mem+24] & 0x00ff) +
			       ((newOB[mem+25]<<8) & 0xff00);
			homeptr = newOT[home/2] & newLocation;
			method = (newOB[homeptr+20] & 0x00ff) +
				 ((newOB[homeptr+21]<<8) & 0xff00);
			methptr = newOT[method/2] & newLocation;
			nlits = newOB[methptr+8] & 0x007e;
			ip = (newOB[mem+8] & 0x00ff) +
			     ((newOB[mem+9]<<8) & 0xff00);
			ip += (nlits+2)<<1;
			newOB[mem+8] = ip & 0xff;
			newOB[mem+9] = (ip>>8) & 0xff;
			ip = (newOB[mem+20] & 0x00ff) +
			     ((newOB[mem+21]<<8) & 0xff00);
			ip += (nlits+2)<<1;
			newOB[mem+20] = ip & 0xff;
			newOB[mem+21] = (ip>>8) & 0xff; }
		mem = newOT[class/2] & newLocation;
		mem += 17;
		if ((newOB[mem] & 0xc0) == 0x40)
			if ((class != oopFloat) & (class != oopBitmap)) {
				printf("Unknown word-type object; ");
				printf("oop = %x\n",class); }}

	if (changes != NULL) {
		printf("\nProcessing changes...\n");
		i = NULL;
		j = 0;
		while (i != EOF) {
		i = fscanf(changes,"oop=%x,field=%x,new=%x ",&oop,&field,&new);
		j += 1;
		if (i == 3) {
			printf("Changing: oop=%x,field=%x,new=%x\n",
				oop,field,new);
			mem = newOT[oop/2] & newLocation;
			mem += (field + 2) * 4;
			for (k = 0; k < 4; k++) {
				temp = new & 0x000000ff;
				new >>= 8;
				newOB[mem++] = temp; }}
		else printf("Input conversion error on line %d\n",j); }}

	fwrite(header, 512, 1, dec);
	fwrite(newOB, 512, (header[0] + 511) >> 9, dec);
	fwrite(newOT, 512, (header[1] + 511) >> 9, dec);
	printf("\n%s successfully created.\n",decFile);
}
